
/*
 * TicTacToe
 *
 * Thomas David Baker, 2002-03-10
 */

package net.bluebones.tictactoe;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

/**
 * Mother class for the <code>net.bluebones.tictactoe</code> package.
 * This class is executable and expects as two commandline arguments
 * the names of the classes of the two players.  The class "Human" is
 * included for human players, as are several AI players.
 *
 * @author  Thomas David Baker, bakert+tictactoe@gmail.com
 * @version 1.1
 */
public class TicTacToe {

    /** <code>byte</code> representing empty. */
    public static final byte EMPTY = -1;
    /** <code>int</code> representing noughts (o) */
    public static final byte NOUGHTS = 0;
    /** <code>int</code> representing crosses (x) */
    public static final byte CROSSES = 1;
    
    /**
     * Controlling method of the <code>TicTacToe</code> executable.  Begins
     * a game of tic-tac-toe.  Expects class names of the players as two
     * commandline arguments.
     *
     * @param   args    <code>String[]</code> commandline arguments.  This 
     *                  should be the class names of the players.  If the 
     *                  players are in the <code>net.blubones.tictactoe</code> 
     *                  package the qualifier can be omitted, otherwise the full
     *                  class names are required.
     */
    public static void main(String[] args) {
        
        Game game = null;
        if (args.length != 2) {
            game = new Game(new Player[] { new Human(), new Human() });
        } else {
            Player[] players = new Player[2];
            for (int i = 0; i < 2; i++) {
                players[i] = getPlayer(args[i]);
            }
            game = new Game(players);
        }
        game.play();
    }
    
    /**
     * Takes supplied integer and returns its "reverse", that is the number of 
     * the opposing player.  That is, 
     * <code>TicTacToe.reverse(TicTacToe.NOUGHTS)</code> returns 
     * <code>TicTacToe.CROSSES</code> and vice versa.  Any other inputs are 
     * illegal and will throw an exception.
     *
     * @param   i   <code>int</code> to "reverse".
     * @return      <code>int</code> of reversed <code>i</code>.
     */
    public static byte reverse(byte i) {

        if (i == TicTacToe.NOUGHTS) {
            return TicTacToe.CROSSES;
        } else if (i == TicTacToe.CROSSES) {
            return TicTacToe.NOUGHTS;
        } else {
            throw new IllegalArgumentException("Number supplied was " + i
                + " but must be TicTacToe.NOUGHTS (" + TicTacToe.NOUGHTS 
                + ") or TicTacToe.CROSSES (" + TicTacToe.CROSSES + ")");
        }
    }
    
    // Gets an instance of a player from its class name.
    private static Player getPlayer(String name) {
            
        String fullName = name;
        
        if (fullName.indexOf(".") == -1) {
            fullName = "net.bluebones.tictactoe." + name;
        }
        
        String errorMsg = "Could not get player because of exception: ";
        
        Object[] arguments;
        Constructor constructor = null;

        try {
            Class c = Class.forName(fullName);
            Class[] arglist = new Class[] {};
            constructor = c.getConstructor(arglist);
        } catch (ClassNotFoundException e) {
            System.err.println(errorMsg + e);
            System.exit(1);
        } catch (NoSuchMethodException e) {
            System.err.println(errorMsg + e);
            System.exit(1);
        }
                
        try {
            arguments = new Object[] {};
            return (Player) constructor.newInstance(arguments);
        } catch (InstantiationException e) {
            System.err.println(errorMsg + e);
            System.exit(1);
        } catch (IllegalAccessException e) {
            System.err.println(errorMsg + e);
            System.exit(1);
        } catch (InvocationTargetException e) {
            System.err.println(errorMsg + e);
            System.exit(1);
        }
        
        return null;
    }
}
